#ifndef mem_h
#define mem_h

#include <stdlib.h>


#ifdef DEBUG_MEM

/**
 * mem_malloc
 *
 * Drop in replacement for malloc, handles logging
 *
 * @param szSize Size of block to allocate
 * @return Pointer to allocated memory or NULL on failure 
 */
extern void *mem_malloc(size_t szSize, const char *sFile, unsigned uLine);

/**
 * mem_calloc
 *
 * Drop in replacement for calloc, handles logging
 *
 * @param sznMeb number of blocks to allocate
 * @param szSize Size of block to allocate
 * @return Pointer to allocated memory or NULL on failure 
 */
extern void *mem_calloc(size_t szNmeb, size_t szSize, const char *sFile, unsigned uLine);

/**
 * mem_free
 *
 * Drop in replacement for free, handles un-logging
 *
 * @param vpMemPtr pointer to memory to free
 */
extern void mem_free(void *vpMemPtr, const char *sFile, unsigned uLine);

/**
 * mem_dump
 *
 * output all unfreed memory blocks
 */
extern void mem_dump(void);

#define MEM_MALLOC(szSize)          mem_malloc(szSize, __FILE__, __LINE__)
#define MEM_CALLOC(szNmeb, szSize)  mem_calloc(szNmeb, szSize, __FILE__, __LINE__)
#define MEM_FREE(vpMemPtr)          mem_free(vpMemPtr, __FILE__, __LINE__)
#else
#define MEM_MALLOC(szSize)          malloc(szSize)
#define MEM_CALLOC(szNmeb, szSize)  calloc(szNmeb, szSize)
#define MEM_FREE(vpMemPtr)          free(vpMemPtr)
#endif



#endif /* mem_h */
